@@ -1,7 +1,6 @@
 import dataclasses
 import json
 import time
-from dataclasses import dataclass
 from enum import Enum
 from typing import List
 from urllib.parse import urlencode, urljoin
@@ -52,7 +51,7 @@ class InvalidStateError(WaldurClientException):
     pass
 
 
-@dataclass
+@dataclasses.dataclass
 class ComponentUsage:
     # TODO: rename to 'component_type' after https://opennode.atlassian.net/browse/WAL-4259 is done
     type: str
@@ -60,7 +59,7 @@ class ComponentUsage:
     description: str = ''
 
 
-@dataclass
+@dataclasses.dataclass
 class ResourceReportRecord:
     header: str
     body: str
@@ -80,36 +79,38 @@ class ProjectRole(Enum):
 
 class WaldurClient(object):
     class Endpoints(object):
-        Provider = 'service-settings'
-        Project = 'projects'
+        ComponentUsage = 'marketplace-component-usages'
+        CustomerPermissions = 'customer-permissions'
+        Customers = 'customers'
         Flavor = 'openstacktenant-flavors'
-        Subnet = 'openstacktenant-subnets'
         FloatingIP = 'openstacktenant-floating-ips'
-        SecurityGroup = 'openstacktenant-security-groups'
         Image = 'openstacktenant-images'
         Instance = 'openstacktenant-instances'
-        Snapshot = 'openstacktenant-snapshots'
-        SshKey = 'keys'
-        Tenant = 'openstack-tenants'
-        TenantSecurityGroup = 'openstack-security-groups'
-        Volume = 'openstacktenant-volumes'
-        VolumeType = 'openstacktenant-volume-types'
+        Invoice = 'invoices'
+        MarketplaceCategories = 'marketplace-categories'
         MarketplaceOffering = 'marketplace-offerings'
-        MarketplacePlan = 'marketplace-plans'
         MarketplaceOrder = 'marketplace-orders'
         MarketplaceOrderItem = 'marketplace-order-items'
+        MarketplacePlan = 'marketplace-plans'
         MarketplaceResources = 'marketplace-resources'
-        MarketplaceCategories = 'marketplace-categories'
-        Customers = 'customers'
-        Invoice = 'invoices'
-        ComponentUsage = 'marketplace-component-usages'
-        RemoteEduteams = 'remote-eduteams'
-        ProjectPermissions = 'project-permissions'
-        CustomerPermissions = 'customer-permissions'
         OfferingPermissions = 'marketplace-offering-permissions'
-        Users = 'users'
-        UserInvitations = 'user-invitations'
         OfferingUsers = 'marketplace-offering-users'
+        Project = 'projects'
+        ProjectPermissions = 'project-permissions'
+        ProjectTypes = 'project-types'
+        Provider = 'service-settings'
+        RemoteEduteams = 'remote-eduteams'
+        SecurityGroup = 'openstacktenant-security-groups'
+        ServiceProviders = 'marketplace-service-providers'
+        Snapshot = 'openstacktenant-snapshots'
+        SshKey = 'keys'
+        Subnet = 'openstacktenant-subnets'
+        Tenant = 'openstack-tenants'
+        TenantSecurityGroup = 'openstack-security-groups'
+        UserInvitations = 'user-invitations'
+        Users = 'users'
+        Volume = 'openstacktenant-volumes'
+        VolumeType = 'openstacktenant-volume-types'
 
     marketplaceScopeEndpoints = {
         'OpenStackTenant.Instance': Endpoints.Instance,
@@ -173,6 +174,8 @@ class WaldurClient(object):
             error = self._parse_error(response)
             raise WaldurClientException(error)
 
+        if method == 'head':
+            return response
         if response.text:
             return response.json()
         return ''
@@ -210,9 +213,16 @@ class WaldurClient(object):
 
         return result
 
+    def _get_count(self, url):
+        response = self._head(url)
+        return int(response.headers['X-Result-Count'])
+
     def _get(self, url, valid_states, **kwargs):
         return self._make_request('get', url, valid_states, 1, **kwargs)
 
+    def _head(self, url):
+        return self._make_request('head', url, valid_states=[200])
+
     def _post(self, url, valid_states, **kwargs):
         return self._make_request('post', url, valid_states, 3, **kwargs)
 
@@ -291,6 +301,8 @@ class WaldurClient(object):
         :return: a resource as a dictionary.
         :raises: WaldurClientException if resource could not be received or response failed.
         """
+        if not value:
+            raise WaldurClientException('Empty ID is not allowed.')
         if is_uuid(value):
             return self._query_resource_by_uuid(endpoint, value, extra)
         else:
@@ -328,6 +340,10 @@ class WaldurClient(object):
         url = self._build_url(self.Endpoints.Users)
         return self._get_all(url)
 
+    def count_users(self):
+        url = self._build_url(self.Endpoints.Users)
+        return self._get_count(url)
+
     def list_ssh_keys(self):
         url = self._build_url(self.Endpoints.SshKey)
         return self._get_all(url)
@@ -461,6 +477,10 @@ class WaldurClient(object):
         resource = self._query_resource_by_uuid(self.Endpoints.Instance, uuid)
         return len(resource['external_ips']) > 0
 
+    def list_tenants(self, filters=None):
+        endpoint = self._build_url(self.Endpoints.Tenant)
+        return self._query_resource_list(endpoint, filters)
+
     def create_security_group(
         self,
         tenant,
@@ -663,14 +683,9 @@ class WaldurClient(object):
 
         return self._query_resource_list(self.Endpoints.MarketplaceResources, params,)
 
-    def marketplace_resource_set_state(self, resource_uuid: str, state: ResourceState):
-        url = self._build_resource_url(
-            self.Endpoints.MarketplaceResources,
-            resource_uuid,
-            action='set_state_by_provider',
-        )
-        payload = {'state': state.value}
-        return self._post(url, valid_states=[200], json=payload)
+    def count_marketplace_resources(self):
+        url = self._build_url(self.Endpoints.MarketplaceResources)
+        return self._get_count(url)
 
     def marketplace_resource_set_backend_id(self, resource_uuid: str, backend_id: str):
         url = self._build_resource_url(
@@ -931,7 +946,7 @@ class WaldurClient(object):
             return self._get_resource(
                 self.Endpoints.MarketplaceOffering,
                 offering,
-                {'project_uuid': project_uuid},
+                {'project_uuid': project_uuid, 'state': ['Active', 'Paused']},
             )
         else:
             return
@@ -969,6 +984,21 @@ class WaldurClient(object):
             WaldurClient.Endpoints.MarketplaceOrderItem, order_item_uuid
         )
 
+    def list_order_items(self, filters=None):
+        return self._query_resource_list(self.Endpoints.MarketplaceOrderItem, filters)
+
+    def marketplace_order_item_approve(self, order_item_uuid: str):
+        url = self._build_resource_url(
+            self.Endpoints.MarketplaceOrderItem, order_item_uuid, action='approve',
+        )
+        return self._post(url, valid_states=[200])
+
+    def marketplace_order_item_reject(self, order_item_uuid: str):
+        url = self._build_resource_url(
+            self.Endpoints.MarketplaceOrderItem, order_item_uuid, action='reject',
+        )
+        return self._post(url, valid_states=[200])
+
     def _get_resource_from_creation_order(
         self, order_uuid, resource_field='resource_uuid', interval=10, timeout=600,
     ):
@@ -1271,27 +1301,31 @@ class WaldurClient(object):
     def list_customers(self, filters=None):
         return self._query_resource_list(self.Endpoints.Customers, filters)
 
+    def count_customers(self):
+        url = self._build_url(self.Endpoints.Customers)
+        return self._get_count(url)
+
     def create_customer(
         self,
         name,
-        email="",
-        address="",
-        registration_code="",
-        backend_id="",
-        abbreviation="",
-        bank_account="",
-        bank_name="",
-        contact_details="",
-        country="",
-        display_name="",
-        domain="",
-        homepage="",
-        native_name="",
+        email='',
+        address='',
+        registration_code='',
+        backend_id='',
+        abbreviation='',
+        bank_account='',
+        bank_name='',
+        contact_details='',
+        country='',
+        display_name='',
+        domain='',
+        homepage='',
+        native_name='',
         latitude=None,
         longitude=None,
-        phone_number="",
-        postal="",
-        vat_code="",
+        phone_number='',
+        postal='',
+        vat_code='',
     ):
         payload = {
             'abbreviation': abbreviation,
@@ -1330,16 +1364,45 @@ class WaldurClient(object):
     def list_projects(self, filters=None):
         return self._query_resource_list(self.Endpoints.Project, filters)
 
-    def create_project(self, customer_uuid, name, backend_id=None):
-        payload = {
+    def count_projects(self):
+        url = self._build_url(self.Endpoints.Project)
+        return self._get_count(url)
+
+    def _serialize_project(
+        self,
+        name=None,
+        backend_id=None,
+        description='',
+        end_date=None,
+        oecd_fos_2007_code=None,
+        type_uuid=None,
+    ):
+        type_url = type_uuid and self._build_resource_url(
+            self.Endpoints.ProjectTypes, type_uuid
+        )
+        return {
             'name': name,
-            'customer': self._build_resource_url(
-                self.Endpoints.Customers, customer_uuid
-            ),
             'backend_id': backend_id,
+            'description': description,
+            'end_date': end_date,
+            'oecd_fos_2007_code': oecd_fos_2007_code,
+            'type': type_url,
         }
+
+    def create_project(self, customer_uuid, name, **kwargs):
+        payload = self._serialize_project(name=name, **kwargs)
+        payload['customer'] = self._build_resource_url(
+            self.Endpoints.Customers, customer_uuid
+        )
+
         return self._create_resource(self.Endpoints.Project, payload=payload)
 
+    def update_project(self, project_uuid, **kwargs):
+        payload = self._serialize_project(**kwargs)
+        return self._update_resource(
+            self.Endpoints.Project, project_uuid, payload=payload
+        )
+
     def delete_project(self, project):
         """
         Delete a project by UUID or URL
@@ -1460,6 +1523,9 @@ class WaldurClient(object):
             self.Endpoints.ProjectPermissions, query_params
         )
 
+    def list_project_permissions(self, filters=None):
+        return self._query_resource_list(self.Endpoints.ProjectPermissions, filters)
+
     def update_project_permission(self, permission_id, new_expiration_time):
         return self._update_resource(
             self.Endpoints.ProjectPermissions,
@@ -1498,6 +1564,9 @@ class WaldurClient(object):
             self.Endpoints.CustomerPermissions, query_params
         )
 
+    def list_customer_permissions(self, filters=None):
+        return self._query_resource_list(self.Endpoints.CustomerPermissions, filters)
+
     def update_customer_permission(self, permission_id, new_expiration_time):
         return self._update_resource(
             self.Endpoints.CustomerPermissions,
@@ -1568,6 +1637,53 @@ class WaldurClient(object):
 
         return self._create_resource(self.Endpoints.OfferingUsers, payload)
 
+    def set_offerings_username(
+        self, service_provider_uuid: str, user_uuid: str, username: str
+    ):
+        endpoint = self._build_resource_url(
+            self.Endpoints.ServiceProviders,
+            service_provider_uuid,
+            'set_offerings_username',
+        )
+        payload = {
+            'user_uuid': user_uuid,
+            'username': username,
+        }
+        return self._post(endpoint, valid_states=[201], json=payload)
+
+    def list_remote_offering_users(self, filters):
+        return self._query_resource_list(self.Endpoints.OfferingUsers, filters)
+
+    def list_service_providers(self, filters):
+        endpoint = self._build_url(self.Endpoints.ServiceProviders)
+        return self._query_resource_list(endpoint, filters)
+
+    def list_service_provider_users(self, service_provider_uuid):
+        endpoint = self._build_resource_url(
+            self.Endpoints.ServiceProviders, service_provider_uuid, 'users'
+        )
+        return self._query_resource_list(endpoint, None)
+
+    def list_service_provider_projects(self, service_provider_uuid):
+        endpoint = self._build_resource_url(
+            self.Endpoints.ServiceProviders, service_provider_uuid, 'projects'
+        )
+        return self._query_resource_list(endpoint, None)
+
+    def list_service_provider_project_permissions(self, service_provider_uuid):
+        endpoint = self._build_resource_url(
+            self.Endpoints.ServiceProviders,
+            service_provider_uuid,
+            'project_permissions',
+        )
+        return self._query_resource_list(endpoint, None)
+
+    def list_service_provider_ssh_keys(self, service_provider_uuid):
+        endpoint = self._build_resource_url(
+            self.Endpoints.ServiceProviders, service_provider_uuid, 'keys'
+        )
+        return self._query_resource_list(endpoint, None)
+
 
 def waldur_full_argument_spec(**kwargs):
     spec = dict(
